package cn.estsh.i3plus.core.apiservice.util;

import cn.estsh.i3plus.core.api.iservice.busi.ISysConfigService;
import cn.estsh.i3plus.platform.common.tool.ReflexTool;
import cn.estsh.i3plus.platform.common.tool.StringTool;
import cn.estsh.i3plus.platform.common.tool.TimeTool;
import cn.estsh.i3plus.platform.common.util.PlatformConstWords;
import cn.estsh.i3plus.pojo.base.annotation.AnnoOutputColumn;
import cn.estsh.i3plus.pojo.base.enumutil.CommonEnumUtil;
import cn.estsh.i3plus.pojo.platform.bean.SysTool;
import cn.estsh.impp.framework.boot.exception.ImppExceptionBuilder;
import cn.estsh.impp.framework.boot.exception.ImppExceptionEnum;
import io.swagger.annotations.ApiParam;
import org.apache.commons.lang3.StringUtils;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.persistence.EntityManager;
import java.io.*;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Ref;
import java.util.*;

/**
 * @Description : Excel导入导出工具类
 * @Reference :
 * @Author : yunhao
 * @CreateDate : 2018-12-05 16:12
 * @Modify:
 **/
@Component
public class ExcelUtil {
    public static final Logger LOGGER = LoggerFactory.getLogger(ExcelUtil.class);

    private ExcelUtil() throws IOException {
    }

    private static ISysConfigService sysConfigService;

    @Autowired
    public void setSysConfigService(ISysConfigService sysConfigService) {
        ExcelUtil.sysConfigService = sysConfigService;
    }

    private static EntityManager entityManager;

    @Autowired
    public void setEntityManager(EntityManager entityManager) {
        ExcelUtil.entityManager = entityManager;
    }


    /**
     * 导出
     *
     * @param data        导出的数据
     * @param exportClass 导出数据类型
     * @param exportCol   需要导出的列
     */
    public static byte[] exportData(List data, Class exportClass, String[] exportCol) {
        File excelFile = null;
        FileOutputStream fos = null;
        //创建HSSFWorkbook对象(excel的文档对象)
        HSSFWorkbook workbook = new HSSFWorkbook();
        try {
            //创建临时文件
            excelFile = File.createTempFile(exportClass.getSimpleName() + new Date().getTime(), ".xls");
            LOGGER.info("临时文件所在的本地路径：" + excelFile.getCanonicalPath());
            fos = new FileOutputStream(excelFile);

            //建立新的sheet对象（excel的表单）
            HSSFSheet sheet = workbook.createSheet(exportClass.getSimpleName());

            // 创建表头
            HSSFRow tableHeader = sheet.createRow(0);

            // 类数据
            Field[] fields = new Field[exportCol.length];

            String colName;

            // 获取字段中文名 优先使用 OutputColumn.name
            for (int i = 0; i < exportCol.length; i++) {
                fields[i] = exportClass.getDeclaredField(exportCol[i]);
                colName = null;

                if (fields[i].isAnnotationPresent(AnnoOutputColumn.class)) {
                    colName = fields[i].getAnnotation(AnnoOutputColumn.class).name();
                }

                if (fields[i].isAnnotationPresent(ApiParam.class) && StringUtils.isBlank(colName)) {
                    colName = fields[i].getAnnotation(ApiParam.class).value();
                }
                tableHeader.createCell(i, CellType.STRING).setCellValue(colName);
            }

            //加载数据至excel对象
            HSSFRow hssfRow;
            Method method;
            AnnoOutputColumn outputColumn;
            Object CellValue;

            for (int i = 0; i < data.size(); i++) {
                hssfRow = sheet.createRow(i + 1);
                for (int j = 0; j < fields.length; j++) {
                    CellValue = exportClass.getDeclaredMethod("get" + StringTool.toUpperCaseFirstOne(fields[j].getName())).invoke(data.get(i));

                    // 判断是否存在引用关系
                    if (fields[j].isAnnotationPresent(AnnoOutputColumn.class)) {
                        outputColumn = fields[j].getAnnotation(AnnoOutputColumn.class);

                        // 判断是否为枚举字段
                        if (outputColumn.refClass().isEnum()) {
                            method = outputColumn.refClass().getDeclaredMethod(
                                    outputColumn.refForeignKey() + "Of" + StringTool.toUpperCaseFirstOne(outputColumn.value()),
                                    outputColumn.refClass().getDeclaredMethod("get"
                                            + StringTool.toUpperCaseFirstOne(outputColumn.refForeignKey())).getReturnType());
                            CellValue = method.invoke(data.get(i), CellValue);
                        } else {
                            selectByProperty(outputColumn.refClass(), outputColumn.value(), outputColumn.refForeignKey(), CellValue);
                        }
                    }

                    hssfRow.createCell(j, CellType.STRING).setCellValue(String.valueOf(CellValue));
                }
            }

            workbook.write(fos);
            fos.flush();
            return file2Byte(excelFile);
        } catch (Exception e) {
            e.printStackTrace();
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.IO_EXCEPTION_FILE.getCode())
                    .setErrorDetail("Excel 导出错误")
                    .build();
        } finally {
            try {
                workbook.close();
                fos.close();
                excelFile.deleteOnExit();
            } catch (IOException e) {
                LOGGER.error("IOException!", e);
            }
        }
    }

    /**
     * 导入
     *
     * @param fileName    文件名
     * @param inputStream 输入流
     * @param importClass 导入类型
     * @return 导入数据集合
     */
    public static <T> List<T> importData(String fileName, InputStream inputStream, Class importClass) {
        List dataList = new ArrayList<T>();
        String fileType = StringTool.getStringFileSuffix(fileName, false);

        try {
            //根据文件类型及文件输入流新建工作簿对象
            Workbook wb = null;
            if (fileType.equals("xls")) {
                wb = new HSSFWorkbook(inputStream);
            } else if (fileType.equals("xlsx")) {
                wb = new XSSFWorkbook(inputStream);
            } else {
                throw ImppExceptionBuilder.newInstance()
                        .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                        .setErrorCode(ImppExceptionEnum.IO_EXCEPTION_FILE.getCode())
                        .setErrorDetail("Excel 您导入的excel格式不正确")
                        .setErrorSolution("请检查导入文件格式")
                        .build();
            }

            // 校验sheet是否超过最大行数
            Sheet sheet = wb.getSheet(importClass.getSimpleName());
            int maxRow = Integer.parseInt(sysConfigService.getSysConfigByCode(PlatformConstWords.EXCEL_IMPORT_MAX_ROW).getConfigValue());
            if (sheet != null && sheet.getLastRowNum() > maxRow) {
                throw ImppExceptionBuilder.newInstance()
                        .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                        .setErrorCode(ImppExceptionEnum.VARIFY_EXCEPTION.getCode())
                        .setErrorDetail("Excel 数据超过最大行数")
                        .setErrorSolution("请检查是否有空行,或分批导入")
                        .build();
            }

            // excel列名与字段名映射
            Map<String, Field> colName = new HashMap<>();
            ApiParam fieldAnno;
            for (Field field : importClass.getDeclaredFields()) {
                if (field.isAnnotationPresent(ApiParam.class)) {
                    fieldAnno = field.getAnnotation(ApiParam.class);
                    colName.put(fieldAnno.value(), field);
                }
            }

            // 创建表头
            Row tableHeader = sheet.getRow(0);
            Field[] fields = new Field[tableHeader.getLastCellNum()];

            for (int i = 0; i < tableHeader.getLastCellNum(); i++) {
                fields[i] = colName.get(tableHeader.getCell(i).getStringCellValue());
            }

            Row row;
            Object obj;
            Object cellValue = null;
            for (int i = 1; i <= sheet.getLastRowNum(); i++) {
                row = sheet.getRow(i);
                obj = importClass.newInstance();
                for (int j = 0; j < fields.length; j++) {
                    row.getCell(j).setCellType(CellType.STRING);
                    if ("".equals(row.getCell(j).getStringCellValue())) {
                        cellValue = null;
                    } else if (fields[j].getType() == String.class) {
                        cellValue = row.getCell(j).getStringCellValue();
                    } else if (fields[j].getType() == Integer.class) {
                        cellValue = Integer.parseInt(row.getCell(j).getStringCellValue());
                    } else if (fields[j].getType() == Long.class) {
                        cellValue = Long.parseLong(row.getCell(j).getStringCellValue());
                    }

                    importClass.getDeclaredMethod("set" + StringTool.toUpperCaseFirstOne(fields[j].getName()), fields[j].getType())
                            .invoke(obj, cellValue);
                }
                dataList.add(obj);
            }

        } catch (IOException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.IO_EXCEPTION.getCode())
                    .setErrorDetail("IO输入输出异常")
                    .build();
        } catch (IllegalAccessException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.REFLEX_EXCEPTION.getCode())
                    .setErrorDetail("无法访问导入类")
                    .setErrorSolution("请检查导入类访问修饰符")
                    .build();
        } catch (InstantiationException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.REFLEX_EXCEPTION.getCode())
                    .setErrorDetail("无法实例化导入类")
                    .setErrorSolution("请检查导入类是拥有无参构造方法")
                    .build();
        } catch (InvocationTargetException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.REFLEX_EXCEPTION.getCode())
                    .setErrorDetail("属性set方法实现错误")
                    .setErrorSolution("请检查属性set方法参数类型是否正确")
                    .build();
        } catch (NoSuchMethodException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.REFLEX_EXCEPTION.getCode())
                    .setErrorDetail("没有找到属性set方法")
                    .setErrorSolution("请检查属性set方法是否存在")
                    .build();
        }
        return dataList;
    }

    /**
     * 导入模板
     *
     * @param exportClass
     */
    public static byte[] importTemplate(Class exportClass) {
        File excelFile = null;
        FileOutputStream fos = null;
        //创建HSSFWorkbook对象(excel的文档对象)
        HSSFWorkbook workbook = new HSSFWorkbook();

        try {
            //创建临时文件
            excelFile = File.createTempFile(exportClass.getSimpleName() + new Date().getTime(), "Tp.xls");
            LOGGER.info("临时文件所在的本地路径：" + excelFile.getCanonicalPath());
            fos = new FileOutputStream(excelFile);

            //建立新的sheet对象（excel的表单）
            HSSFSheet sheet = workbook.createSheet(exportClass.getSimpleName());

            // 创建表头
            HSSFRow tableHeader = sheet.createRow(0);
            HSSFRow tableData = sheet.createRow(1);

            // 类数据
            Field[] declaredFields = exportClass.getDeclaredFields();
            ApiParam fieldAnno;
            int col = 0;

            for (int i = 0; i < declaredFields.length; i++) {
                if (declaredFields[i].isAnnotationPresent(ApiParam.class)) {
                    fieldAnno = declaredFields[i].getAnnotation(ApiParam.class);
                    if (!fieldAnno.hidden()) {
                        tableHeader.createCell(col, CellType.STRING).setCellValue(fieldAnno.value());

                        tableData.createCell(col, CellType.STRING).setCellValue(fieldAnno.example());
                        col++;
                    }
                }
            }

            workbook.write(fos);
            return file2Byte(excelFile);
        } catch (IOException e) {
            throw ImppExceptionBuilder.newInstance()
                    .setSystemID(CommonEnumUtil.SOFT_TYPE.CORE.getCode())
                    .setErrorCode(ImppExceptionEnum.IO_EXCEPTION.getCode())
                    .setErrorDetail("Excel导出错误")
                    .build();
        } finally {
            try {
                workbook.close();
                fos.close();
                excelFile.deleteOnExit();
            } catch (IOException e) {
                LOGGER.error("IOException!", e);
            }
        }
    }

    /**
     * 获取实体列集合
     *
     * @param pojoClass
     * @return
     */
    public static Map getColName(Class pojoClass) {
        Map colName = new LinkedHashMap();
        Field[] field = pojoClass.getDeclaredFields();
        ApiParam fieldAnno;

        for (int i = 0; i < field.length; i++) {
            if (field[i].isAnnotationPresent(ApiParam.class)) {
                fieldAnno = field[i].getAnnotation(ApiParam.class);
                if (!fieldAnno.hidden()) {
                    colName.put(field[i].getAnnotation(ApiParam.class).value(), field[i].getName());
                }
            }
        }

        return colName;
    }

    /**
     * @param persistentClass
     * @param colName
     * @param propertyName
     * @param value
     * @return
     */
    public static Object selectByProperty(Class persistentClass, String colName, String propertyName, Object value) throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        String queryString = "select model." + colName + " from " + persistentClass.getSimpleName()
                + " as model where model." + propertyName + "= :" + propertyName;
        Object result = entityManager.createQuery(queryString).setParameter(propertyName, value).getSingleResult();

        return persistentClass.getDeclaredMethod("get" + StringTool.toUpperCaseFirstOne(colName)).invoke(result);
    }

    public static void main(String[] args) {
        List<SysTool> sysTools = new ArrayList<>();
        SysTool sysTool = new SysTool();
        for (int i = 0; i < 14; i++) {
            sysTool.setName("ddd");
            sysTools.add(sysTool);
        }
        String[] colName = new String[]{"name", "toolTypeNameRdd", "toolStatus", "toolIp", "toolPort", "toolConnType", "toolDataType", "toolOperating", "toolDescription"};
        try {
            FileOutputStream ds = new FileOutputStream("E://testOut.xls");
            ds.write(ExcelUtil.exportData(sysTools, SysTool.class, colName));
            ds.flush();
            ds.close();
//            ExcelUtil.importData("testOut.xls", new FileInputStream("E://testOut.xls"), SysTool.class);
//            ExcelUtil.importTemplate(new FileOutputStream("E://testOut.xls"), SysTool.class, colName);

        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    public static byte[] file2Byte(File file) {
        byte[] buffer = null;
        try {
            FileInputStream fis = new FileInputStream(file);
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            byte[] b = new byte[1024];
            int n;
            while ((n = fis.read(b)) != -1) {
                bos.write(b, 0, n);
            }
            fis.close();
            bos.close();
            buffer = bos.toByteArray();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return buffer;
    }
}
